/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Channel_Serial.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_CHANNEL_SERIAL_H
#define DCDT_CHANNEL_SERIAL_H

#include <DCDT_Channel.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <iostream>
#include <sys/time.h>
#include <DCDT_ctrace.h>

#define DEVICE_NAME_LEN UNIX_PATH_MAX
#define BAUDRATE B4000000

#define SER_LOST_TIMER_VAL 10000000
#define SER_RECV_TIMER_VAL   100000

class CommData_Serial;

class DCDT_Channel_Serial : public DCDT_Channel {
 public:
  DCDT_Channel_Serial(int AgoraID);
  virtual ~DCDT_Channel_Serial();

  inline void Open(int link_status = 0);
  inline void Reopen(int link_status);
  // the Prepare function is responsible for the allocated area of the two classes passed as parameter
  inline void Prepare(CommData *local, CommData *remote);
  inline void StaticPrepare(unsigned int prof, char *dev);
  inline void ChooseChannel(int channel){channel = 0;};
  inline void WaitConn();
  inline void StartConn();
  inline void Send(const DCDT_Msg *msg);
  inline void SendNotify();
  inline void SendAnswer(CommData* to, CommData* local);
  inline DCDT_Msg* Receive();
  inline void ReceiveHS(HSMsgHeader*& header, CommData*& remote);
  inline void Dispose();
  inline void Close();
  inline void KeepAlive(){};
  /** Returns an heap memory area of wich is no more responsible */
  inline CommData* GetStartingCD();
  inline unsigned int ReadStartingProfile();
  inline CommData* GetCommData();
  inline void SetTimers();
  inline void SetLostTimer(int usec);
  inline void LostTimerOn();
  inline void LostTimerOff();
  inline void SetConnTimer(int usec);
  inline void ConnTimerOn();
  inline void ConnTimerOff();
  inline void SetSendTimer(int usec);
  inline void SendTimerOn();
  inline void SendTimerOff();
  inline void UnblockSend();
  inline void SetReceiveTimer(int usec);
  inline void ReceiveTimerOn();
  inline void ReceiveTimerOff();
  inline void UnblockReceive();
  inline void Restart(int link_status);
  inline void Stop();

 private:
  int device, status;
  CommData_Serial *localCD;
  struct termios options;
  int 
    /// Flag: = 1 -> use timer when receiving
    trecv_flag, 
    /// Flag: = 1 -> use timer when communicating
    tlost_flag, 
    /// Flag: = 1 -> use no blocking calls when receiving
    unblock_recv_flag;

  struct timeval 
    /// Receive timeout
    trecv_value,
    /// Communication timeout
    tlost_value;

  char device_name[UNIX_PATH_MAX];
  unsigned int profile;

  /// Flag: 1 -> dynamic channel; != 1 -> static channel
  bool dynamic;
};

class CommData_Serial : public CommData {
public:
  char *device_name;

  CommData_Serial();
  CommData_Serial(char *name);
  CommData_Serial(CommData_Serial *cd);
  virtual ~CommData_Serial();

  CommData_Serial& operator=(const CommData_Serial& cd) {
    strcpy(device_name, cd.device_name);
    return (*this);
  }

  DCDT_Channel* CreateChannel(int AgoraID) {
    return (new DCDT_Channel_Serial(AgoraID));
  }
};

/*! \brief Prepare channel's data structure.
 */
inline void DCDT_Channel_Serial::Prepare(CommData *local, CommData *remote)
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Prepare channel Serial"));
  if (local) {
    if (localCD)
      delete localCD;
    localCD = (CommData_Serial*)local;
    
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("localCD: device_name=%s", localCD->device_name));
  } else
    localCD = new CommData_Serial();// (???)
  strcpy(device_name, localCD->device_name);

  remote = NULL;
}

/*! \brief Prepare static channel's data structure, open a FD. 
 */
inline void DCDT_Channel_Serial::StaticPrepare(unsigned int prof, char *dev)
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("StaticPrepare channel Serial"));
  dynamic = false;
  profile = prof;

  localCD = new CommData_Serial(dev);
  strcpy(device_name, dev);
  Open();
  SetReceiveTimer(0);
  ReceiveTimerOn();
  SetLostTimer(SER_LOST_TIMER_VAL);
  LostTimerOn();
}

/*! \brief Try to open a serial device.
 *
 *  \exception ChannelError Unable to open the specified device
 */
inline void DCDT_Channel_Serial::Open(int link_status )
{
  device = open(localCD->device_name, O_RDWR | O_NOCTTY | O_NDELAY);
  if (device == -1) {
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ( "Unable to open %s",  localCD->device_name));
    throw ChannelError();
  }
  tcgetattr(device, &options);

  // Set serial channel parameters
  options.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
  options.c_iflag = 0;
  options.c_oflag = 0;
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG );

  tcsetattr(device, TCSAFLUSH, &options);
  fcntl(device, F_SETFL, 0);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial port opened: %s", localCD->device_name));
}

inline void DCDT_Channel_Serial::Reopen(int link_status)
{
  Open(link_status);
}

inline CommData* DCDT_Channel_Serial::GetCommData()
{
  return localCD;
}

/*! \brief Method called when the link is in WAITING state.
 *
 *  Wait for an HS_NOTIFY message then respond with an HS_ANSWER message.
 *  ( ??? ) indefinitamente
 */
inline void DCDT_Channel_Serial::WaitConn()
{
  HSMsgHeader *h;
  CommData *c;
  ReceiveTimerOff();
  ReceiveHS(h, c);
  SendNotify();
}

/*! \brief Receive a message using the channel settings.
 *
 *  \exception ConnError
 *  \return NULL No new message received\n
 *  != NULL Pointer to the message received
 */
inline DCDT_Msg* DCDT_Channel_Serial::Receive()
{
  DCDT_MsgHeader *msg_header;
  DCDT_Msg *msg;
  int plen;
  void *msg_payload;

  char *index;
  int nread = 0, toread = MSGHEADER_LEN;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  FD_ZERO(&rset);
  FD_SET(device, &rset);
  if (unblock_recv_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (trecv_flag)
  {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(device+1, &rset, NULL, NULL, tv_ptr))
  {
    case 0: return NULL;
    case -1: throw ConnError(errno);
  }
  // Start receiving header
  msg_header = new DCDT_MsgHeader;
  index = (char*)msg_header;
  if ((nread = read(device, index, toread)) < 0)
  {
    delete msg_header;
    msg_header = NULL;
    throw ConnError(errno);
  }
  else if (nread == 0)
  {
    delete msg_header;
    msg_header = NULL;
    throw ConnError();
  }
  index += nread;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial header byte read: %i", nread));

  toread -= nread;

  while ( toread )
  {
    FD_ZERO(&rset);
    FD_SET(device, &rset);
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(device+1, &rset, NULL, NULL, tv_ptr))
    {
      case 0:
      case -1:
        delete msg_header;
        msg_header = NULL;
        throw ConnError(errno);
    }
    if ((nread = read(device, index, toread)) < 0)
    {
      delete msg_header;
      msg_header = NULL;
      throw ConnError(errno);
    }
    else if (nread == 0)
    {
      delete msg_header;
      msg_header = NULL;
      throw ConnError();
    }
    index += nread;

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial header byte read: %i", nread));

    toread -= nread;
  }
  msg = new DCDT_Msg(msg_header);

  // Receivig payload
  msg_payload = malloc(plen = msg->ReadPayloadLen());
  index = (char*)msg_payload;
  toread = plen;
  while ( toread )
  {
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(device+1, &rset, NULL, NULL, tv_ptr))
    {
      case 0:
      case -1: throw ConnError(errno);
    }
    if ((nread = read(device, index, toread)) < 0)
    {
      free(msg_payload);
      delete msg_header;
      msg_header = NULL;
      throw ConnError(errno);
    }
    else if (nread == 0)
    {
      free(msg_payload);
      delete msg_header;
      msg_header = NULL;
      throw ConnError();
    }
    index += nread;

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial payload byte read: %i", nread));
    toread -= nread;
  }
  msg->SetPayload(msg_payload, plen);

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial message successfully received"));

  return msg;
}

/*! \brief Receive an handshake message.
 *
 *  When an HS_ANSWER message was received close and reopen the channel.
 *  \exception ConnError
 */
inline void DCDT_Channel_Serial::ReceiveHS(HSMsgHeader*& header, CommData*& remote)
{
  char *index;
  int nread = 0, toread = HSMSGHEADER_LEN;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  FD_ZERO(&rset);
  FD_SET(device, &rset);

  if (trecv_flag)
  {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(device+1, &rset, NULL, NULL, tv_ptr))
  {
    case 0: return;
    case -1: throw ConnError(errno);
  }

  // Start receiving header
  header = new HSMsgHeader();
  index = (char*)header;
  if ((nread = read(device, index, toread)) < 0)
    {
      //      cout << "\nread:" << nread;
      delete header;
      header = NULL;
      throw ConnError(errno);
    }
  else if (nread == 0)
    {
      delete header;
      header = NULL;
      throw ConnError();
    }
  index += nread;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial header byte read: %i", nread));

  toread -= nread;

  while ( toread )
  {
    FD_ZERO(&rset);
    FD_SET(device, &rset);
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(device+1, &rset, NULL, NULL, tv_ptr))
    {
      case 0:
      case -1:
        delete header;
        header = NULL;
        throw ConnError(errno);
    }
    if ((nread = read(device, index, toread)) < 0)
    {
      delete header;
      header = NULL;
      throw ConnError(errno);
    }
    else if (nread == 0)
    {
      delete header;
      header = NULL;
      throw ConnError();
    }
    index += nread;

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial header byte read: %i", nread));

    toread -= nread;
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial h-s message successfully received"));
  if (header->type == HS_ANSWER) {
    Stop();
    remote = new CommData_Serial(device_name);
  }
}

/*! \brief Send a message.
 *
 *  \exception TimeOut Timeout waiting for the channel
 *  \exception ConnError
 */
inline void DCDT_Channel_Serial::Send(const DCDT_Msg *msg)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  int sent, tosend = MSGHEADER_LEN;
  char *sending_ptr;

  // Sending Header
  sending_ptr = (char*)msg->GetHeader();
  while (tosend) {
    FD_ZERO(&wset);
    FD_SET(device, &wset);
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(device+1, NULL, &wset, NULL, tv_ptr))
    {
      case 0: throw TimeOut();
      case -1: throw ConnError(errno);
    }

    if ((sent = write(device, sending_ptr, tosend)) < 0)
      throw ConnError();

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial header byte sent: %i", sent));

    tosend -= sent;
    sending_ptr += sent;
  }

  // Sending payload
  sending_ptr = (char*)msg->GetPayload();
  tosend = msg->ReadPayloadLen();
  while (tosend) {
    FD_ZERO(&wset);
    FD_SET(device, &wset);
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(device+1, NULL, &wset, NULL, tv_ptr))
    {
      case 0:
      case -1: throw ConnError(errno);
    }

    if ((sent = write(device, sending_ptr, tosend)) < 0)
      throw ConnError();

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial payload byte sent: %i", sent));

    tosend -= sent;
    sending_ptr += sent;
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial message successfully sent"));
}

/*! \brief Send an HS_NOTIFY message.
 *
 *  \exception ConnError
 */
inline void DCDT_Channel_Serial::SendNotify()
{
  int sent, tosend = HSMSGHEADER_LEN;
  char *sending_ptr;

  // Sending
  sending_ptr = (char*)&notify;
  while (tosend) {
    sent = write(device, sending_ptr, tosend);
    if (sent < 0)
      throw ConnError();

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial Notify byte sent: %i", sent));

    tosend -= sent;
    sending_ptr += sent;
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial Notify successfully sent"));
}

/*! \brief Send an HS_ANSWER message then stop the channel.
 *
 *  \exception ConnError
 */
inline void DCDT_Channel_Serial::SendAnswer(CommData* to, CommData* local)
{
  int sent, tosend = HSMSGHEADER_LEN;
  char *sending_ptr;

  // SENDING
  sending_ptr = (char*)&answer;
  while (tosend) {
    if ((sent = write(device, sending_ptr, tosend)) < 0)
      throw ConnError();

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial Answer byte sent: %i", sent));

    tosend -= sent;
    sending_ptr += sent;
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Serial Answer successfully sent"));
  Stop();

  to = NULL;
  local = NULL;
}

inline CommData* DCDT_Channel_Serial::GetStartingCD()
{
  CommData_Serial *cd = new CommData_Serial(device_name);
  return cd;
}

inline unsigned int DCDT_Channel_Serial::ReadStartingProfile()
{
  return profile;
}

/*! \brief Method called when the link is in CONNECTING state.
 */
inline void DCDT_Channel_Serial::StartConn()
{
  HSMsgHeader *h;
  CommData *c;
  SendNotify();
  ReceiveTimerOff();
  ReceiveHS(h, c);
}

inline void DCDT_Channel_Serial::Dispose()
{
}

inline void DCDT_Channel_Serial::Close()
{
  if (device) {
    close(device);
    device = 0;
  }
}

inline void DCDT_Channel_Serial::SetTimers()
{
  SetReceiveTimer(SER_RECV_TIMER_VAL);
  ReceiveTimerOn();
  SetLostTimer(SER_LOST_TIMER_VAL);
  LostTimerOn();
}

inline void DCDT_Channel_Serial::SetLostTimer(int usec)
{
  tlost_value.tv_usec = usec % 1000000;
  tlost_value.tv_sec = usec / 1000000;
}

inline void DCDT_Channel_Serial::LostTimerOn()
{
  tlost_flag = 1;
}

inline void DCDT_Channel_Serial::LostTimerOff()
{
  tlost_flag = 0;
}

inline void DCDT_Channel_Serial::SetConnTimer(int usec)
{
  usec = 0;
}

inline void DCDT_Channel_Serial::ConnTimerOn()
{
}

inline void DCDT_Channel_Serial::ConnTimerOff()
{
}

inline void DCDT_Channel_Serial::SetSendTimer(int usec)
{
  usec = 0;
}

inline void DCDT_Channel_Serial::SendTimerOn()
{
}

inline void DCDT_Channel_Serial::SendTimerOff()
{
}

inline void DCDT_Channel_Serial::UnblockSend()
{
}

inline void DCDT_Channel_Serial::SetReceiveTimer(int usec)
{
  trecv_value.tv_usec = usec % 1000000;
  trecv_value.tv_sec = usec / 1000000;
}

inline void DCDT_Channel_Serial::ReceiveTimerOn()
{
  trecv_flag = 1;
  unblock_recv_flag = 0;
}

inline void DCDT_Channel_Serial::ReceiveTimerOff()
{
  trecv_flag = 0;
}

inline void DCDT_Channel_Serial::UnblockReceive()
{
  unblock_recv_flag = 1;
}

inline void DCDT_Channel_Serial::Restart(int link_status)
{
  Open(link_status);
}

inline void DCDT_Channel_Serial::Stop()
{
  if (device) {
    close(device);
    device = 0;
  }
}

#endif
