/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_TCPSocket.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef TCPSOCKET_H
#define TCPSOCKET_H

#include <Socket.h>
#include <netdb.h>
#include <netinet/tcp.h>

class TCPSocket : public Socket
{
public:
  TCPSocket() {};
  ~TCPSocket() {};

  inline void Open(int fd = 0);
  inline void Send(const DCDT_Msg* msg);
  inline DCDT_Msg* Receive();
  inline void Listen(int qlen = 5);
  inline TCPSocket* Accept();
  inline void SetBufs(int rcvbuf, int sndbuf, int rcvlowat, int sndlowat);
  inline void SetSegSize(int size);

};

/*! \brief Open a new tcp socket or use a previously opened one.
 *
 *  If fd!=0, a valid socket descriptor was passed, get the informations about the remote and set connected=1.\n
 *  \exception ChannelError
 */
inline void TCPSocket::Open(int fd)
{
  struct sockaddr_in *sa=NULL;
  socklen_t salen = 0;
  if (opened)
    return;
  if (fd) {
    sockfd = fd;
    if (getpeername(sockfd, (struct sockaddr *) sa, &salen) < 0)
      throw ChannelError(errno);
    connected = 1;
  }
  else if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    throw ChannelError(errno);
  opened = 1;
  
  //if (fcntl(sockfd, F_SETFL, FNDELAY) < 0)
  //  cout << "\nWarning: can't set socket to unblocking";
    
  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCPSocket successfully opened"));
}

/*! \brief Send a message using a connection-orientend protocol ( tcp )
 *
 *  Send the message \a msg in two steps:
 *  -# Send the header
 *  -# Send the payload
 *  .
 *  There are no limits on the legth of both header and payload. Timers precedence: unlock, send, lost.
 *  \exception ChannelError
 *  \exception TimeOut
 */
inline void TCPSocket::Send(const DCDT_Msg* msg)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  int sent, tosend = MSGHEADER_LEN;
  char *sending_ptr;

  if (!opened || !connected)
    return;

  // Sending HEADER
  sending_ptr = (char*)msg->GetHeader();
  while (tosend) {
    FD_ZERO(&wset);
    FD_SET(sockfd, &wset);
    if (unblock_send_flag) {
      tv.tv_sec = 0;
      tv.tv_usec = 0;
      tv_ptr = &tv;
    }
    else if (tsend_flag) {
      tv.tv_sec = tsend_value.tv_sec;
      tv.tv_usec = tsend_value.tv_usec;
      tv_ptr = &tv;
    }
    else if (tlost_flag) {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr)) {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
    }
    
    if ((sent = write(sockfd, sending_ptr, tosend)) < 0)
      throw ConnError();
    
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP header byte sent: %i", sent));
    
    tosend -= sent;
    sending_ptr += sent;
  }
  
  // Sending PAYLOAD
  sending_ptr = (char*)msg->GetPayload();
  tosend = msg->ReadPayloadLen();

  while (tosend) {
    FD_ZERO(&wset);
    FD_SET(sockfd, &wset);
    if (tlost_flag) {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr)) {
    case 0:
    case -1: throw ConnError(errno);
    }
    
    if ((sent = write(sockfd, sending_ptr, tosend)) < 0)
      throw ConnError();
    
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP payload byte sent: %i", sent));
    
    tosend -= sent;
    sending_ptr += sent;
  }
  
}

/*! \brief Receive a message
 *
 *  Such as in Send receive a message in two steps, first the header then the payload; a new DCDT_Msg object is created and returned when a message was received, is up to the user to delete the object.
 *  Timers precedence: unblock, receive, lost.
 *  \exception ConnErr
 *  \return NULL - No message was received
 *  \return !=NULL - Pointer to the recived message, is up to the user to delete the returned object.
 */
inline DCDT_Msg* TCPSocket::Receive()
{
  DCDT_MsgHeader *msg_header;
  DCDT_Msg *msg;
  int plen;
  void *msg_payload;
  char *index;
  int nread = 0, toread = MSGHEADER_LEN;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  if (!opened || !connected)
    return NULL;

  // Receiving header
  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  if (unblock_recv_flag)
  {
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("unblock_recv_flag"));
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (trecv_flag)
  {
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("trecv_flag"));
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("tlost_flag"));
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr))
  {
    case 0: return NULL;
    case -1:
       TRC_PRINT( DCDT_TRC_COMM, TRC0, ("select=-1 -> errno %d", errno));
       throw ConnError(errno);
       break;
  }

  // Start receiving header
  msg_header = new DCDT_MsgHeader;
  index = (char*)msg_header;
  nread = read( sockfd, index, toread );
  msg_header->mtx = new DCDT_Mutex;
  if ( nread< 0)
  {
    delete msg_header;
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header nread=%d - errno=%d", nread, errno));
    throw ConnError(errno);
  }

  /* ???
  else if (nread == 0)
  {
    delete msg_header;
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header nread==0"));
    throw ConnError();
  } */
  index += nread;

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header TCP header byte read: %i", nread));

  toread -= nread;

  while ( toread )
  {
    FD_ZERO(&rset);
    FD_SET(sockfd, &rset);
    if (tlost_flag)
    {
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr))
    {
      case 0:
         TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header2 select=0 -> errno %d", errno));
         //throw ConnError(errno); ???
	 return NULL;
      case -1:
         TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header2 select=-1 -> errno %d", errno));
         throw ConnError(errno);
    }
    if ((nread = read(sockfd, index, toread)) < 0)
    {
      delete msg_header;
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header2 nread=%d -> errno %d", nread, errno));
      throw ConnError(errno);
    }
    else if (nread == 0)
    {
      delete msg_header;
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Header2 nread=0 -> errno %d", errno));
      //throw ConnError(); ???
      return NULL;
    }

    index += nread;

    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP header byte read: %i", nread));

    toread -= nread;
  }
  msg = new DCDT_Msg(msg_header);

  // Receiving payload
  msg_payload = malloc(plen = msg->ReadPayloadLen());
  index = (char*)msg_payload;
  toread = plen;
  while ( toread )
  {
    if (trecv_flag)
    {
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("trecv_flag"));
      tv.tv_sec = trecv_value.tv_sec;
      tv.tv_usec = trecv_value.tv_usec;
      tv_ptr = &tv;
    }
    else if (tlost_flag)
    {
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("tlost_flag"));
      tv.tv_sec = tlost_value.tv_sec;
      tv.tv_usec = tlost_value.tv_usec;
      tv_ptr = &tv;
    }
    switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr))
    {
      case 0:
         TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Payload select=0 -> errno %d", errno));
	 free(msg_payload);
	 delete msg_header;
         //throw ConnError(errno); ???
	 return NULL;
      case -1:
         TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Payload select=-1 -> errno %d", errno));
	 free(msg_payload);
	 delete msg_header;
         throw ConnError(errno);
    }
    if ((nread = read(sockfd, index, toread)) < 0)
    {
      free(msg_payload);
      delete msg_header;
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Payload nread=%d -> errno %d", nread, errno));
      throw ConnError(errno);
    }
    else if (nread == 0)
    {
      free(msg_payload);
      delete msg_header;
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Payload nread=0 -> errno %d", errno));
      //throw ConnError(); ???
      return NULL;
    }
    index += nread;

    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP payload byte read: %i", nread));
    toread -= nread;
  }
  msg->SetPayload(msg_payload, plen);

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP message successfully received"));

  return msg;
}

/*! \brief Listen for connections
 *
 * \exception ChannelError
 */
inline void TCPSocket::Listen(int qlen)
{
  if (!opened)
    return;
  if (listen(sockfd, qlen) < 0)
    throw ChannelError(errno);

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP socket listening"));
}

/*! \brief Accept a new connection.
 * 
 *  Used only the \a conn timer.
 *  \exception TimeOut No connection request in queue
 *  \exception ConnError
 *  \exception ChannelError
 *  \return NULL - No connection request in queue
 *  \return !=NULL - Pointer to the TCP socket connected to a remote socket
 */
inline TCPSocket*
TCPSocket::Accept()
{
  int conn, addrlen;
  struct sockaddr_in sinremote;
  TCPSocket *s=NULL;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  if (!opened)
    return NULL;

  TRC_PRINT( DCDT_TRC_COMM, TRC0, ("TCP socket accepting"));

  addrlen = sizeof(sinremote);

  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  if (tconn_flag)
  {
    tv.tv_sec = tconn_value.tv_sec;
    tv.tv_usec = tconn_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  if ((conn = accept(sockfd, (struct sockaddr*) &sinremote, (socklen_t*) &addrlen)) < 0)
    throw ConnError(errno);
  try
  {
    s = new TCPSocket();
    s->Open(conn);
  }
  catch (ChannelError e)
  {
    delete s;
    throw;
  }
  if (tconn_flag) {
    s->SetConnTimer(tconn_value.tv_sec + tconn_value.tv_usec);
    s->ConnTimerOn();
  }
  if (tlost_flag) {
    s->SetLostTimer(tlost_value.tv_sec + tlost_value.tv_usec);
    s->LostTimerOn();
  }
  if (trecv_flag) {
    s->SetReceiveTimer(trecv_value.tv_sec + trecv_value.tv_usec);
    s->ReceiveTimerOn();
  }
  if (tsend_flag) {
    s->SetSendTimer(tsend_value.tv_sec + tsend_value.tv_usec);
    s->SendTimerOn();
  }
  return s;
}

inline void TCPSocket::SetBufs(int rcvbuf, int sndbuf, int rcvlowat, int sndlowat)
{
  GetOption(SOL_SOCKET, SO_RCVBUF);
  if (rcvbuf) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &rcvbuf, sizeof(rcvbuf)) < 0)
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Warning: can't set receive buffer for tcp socket"));
  }
  GetOption(SOL_SOCKET, SO_RCVBUF);
  GetOption(SOL_SOCKET, SO_SNDBUF);
  if (sndbuf) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &sndbuf, sizeof(sndbuf)) < 0)
      TRC_PRINT(DCDT_TRC_COMM, TRC0, ("Warning: can't set send buffer for tcp socket"));
  }
  GetOption(SOL_SOCKET, SO_SNDBUF);
  GetOption(SOL_SOCKET, SO_RCVLOWAT);
  if (rcvlowat) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVLOWAT, &rcvlowat, sizeof(rcvlowat)) < 0)
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Warning: can't set receive low water mark for tcp socket"));
  }
  GetOption(SOL_SOCKET, SO_RCVLOWAT);
  GetOption(SOL_SOCKET, SO_SNDLOWAT);
  if (sndlowat) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDLOWAT, &sndlowat, sizeof(sndlowat)) < 0)
      TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Warning: can't set send low water mark for tcp socket"));
  }
  GetOption(SOL_SOCKET, SO_SNDLOWAT);
}

inline void TCPSocket::SetSegSize(int size)
{
  GetOption(IPPROTO_TCP, TCP_MAXSEG);
  if (setsockopt(sockfd, IPPROTO_TCP, TCP_MAXSEG, &size, sizeof(size)) < 0)
    TRC_PRINT( DCDT_TRC_COMM, TRC0, ("Warning: can't set maximum segment size for tcp socket"));
  GetOption(IPPROTO_TCP, TCP_MAXSEG);
}

#endif
