/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Member.cpp

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/


#ifndef DCDT_MEMBER_CPP
#define DCDT_MEMBER_CPP

#include <DCDT_Member.h>
#include <DCDT_Agora.h>
#include <DCDT_Finder.h>

#ifdef USE_BOOST
/*
  Boost version
*/
struct BoostThread {
  BoostThread( DCDT_Member *M ) { member = M; }
  DCDT_Member *member;
  void operator()() {
    member->LifeCycle();
  }
};

static void ChangeThreadScheduler ( void )
{

}

#else
#include <string.h> // for strerror_r
/*
  pthread version
*/
extern "C" void *ThreadStartRoutine( void * arg)
{
  DCDT_Member* M = (DCDT_Member *) arg;
  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("ThreadStartRoutine Invoked for Member %d", M->ReadID()));

  // Hook to the method that manages DCDT_Member
  // Needed to access DCDT_Member protected part
  // TODO: review
  return M->LifeCycle();
};

/*! \brief Change scheduler and priority for the current thread
 *
 *  The behavior of this function is controlled by 2 environment variables,
 *  DCDT_SCHED_RR and DCDT_SCHED_PRIO.
 *  If DCDT_SCHED_RR is defined the thread will try to use the realtime scheduler ( SCHED_RR ) with the minimun priority ( 1 )
 *  If DCDT_SCHED_PRIO is definied his value will be used as priority for the scheduler
 */
static void ChangeThreadScheduler ( void )
{
  // Change the thread scheduler and priority if nedded
  int scheduler = SCHED_OTHER, ret;
  struct sched_param sparam;
  
  if ( getenv ( "DCDT_SCHED_PRIO" ) != NULL )
    sparam.sched_priority = atoi(getenv( "DCDT_SCHED_PRIO" ));
  else
    sparam.sched_priority = -1;
  
  if ( getenv ( "DCDT_SCHED_RR" ) != NULL ) {
    scheduler = SCHED_RR;
    sparam.sched_priority = 1;
  }

  if ( ( scheduler != SCHED_OTHER ) || ( sparam.sched_priority != -1 ) ) {
    ret = pthread_setschedparam( pthread_self(), scheduler,  &sparam );
    if ( ret != 0 ) {
      char buf[80];
      printf ( "Error: pthread_setschedparam ( S=%d, P=%d ) = %s (%d)\n", scheduler,
	       sparam.sched_priority, strerror_r(ret, buf, 80), ret );
    }
  }
}

#endif


static void InstallSignalHandler ( void )
{
   struct sigaction sa;
   __sigset_t sigset;
                                                                                
   // Ignore SIGPIPE from sockets
   sa.sa_handler = SIG_IGN;
   memset(&sigset,0,sizeof(sigset));;
   sa.sa_mask = sigset;
   sa.sa_flags = 0;
   sa.sa_restorer = NULL;
   if ( sigaction(SIGPIPE,&sa,NULL) != 0 ) {
      printf ( "Error: Unable to ignore SIGPIPE\n" );
      exit ( 1 );
   }
}

DCDT_Member::DCDT_Member(DCDT_Agora *agora, DCDT_TIME period ) {
  myAgora = agora;

  myAgora->ReadIPStrAddr(myIPAddr);
  ExecTime = 0;
  Period = period;
  CycleTimeExceded=false;
};

DCDT_Member::DCDT_Member(DCDT_Agora *agora) {
  myAgora = agora;
  
  myAgora->ReadIPStrAddr(myIPAddr);
  ExecTime = 0;
  Period = 0;
  CycleTimeExceded=false;
};

DCDT_Member::~DCDT_Member()
{
};

/*! \brief Create a new thread that will execute the member
 */
int DCDT_Member::Activate( const char *thd_name ) {

  if ( thd_name == NULL )
    thd_name = typeid(*this).name();

#ifdef USE_BOOST

  myBoostThread = new BoostThread( this );
  myBoostThd = (void *)new boost::thread( *myBoostThread );

#else
  int ret_crea;
  ret_crea = pthread_create( &myThread, NULL, \
			     ThreadStartRoutine, (void *) this );
  
  if ( ret_crea ) {
    TRC_ERROR ( DCDT_TRC_MEMBER, ( "Cannot create DCDT_Member thread %d '%s'", myID, thd_name ) );
    return FAILURE;
  }
  else {
    TRC_PRINT ( DCDT_TRC_MEMBER, TRC0, ( "Member thread %d '%s' created", myID, thd_name ));
  };

  TRC_ADD_THREAD ( thd_name, myThread );

#endif

  return SUCCESS;
};

/*! \brief Member main function executed by the member's thread.
 *
 *  This functions implements the basic member behavior.
 *  1 - Call Init()
 *  2 - Wait until the Agora' is DCDT_LETSWORK state
 *  3 - Execute the periodical function DoYourJob
 *  4 - Call Close()
 *  5 - Terminate thread
 */
inline void * DCDT_Member::LifeCycle() {

  ChangeThreadScheduler();
  InstallSignalHandler();

  StartTime = GetTime();

  myAgora->SetActive( myID );

  // Member initialization
  Status = INITIALIZING;
  Init();
  
  // Now we are ready
  Status = READY;
  
  // When the Agora is working we can start
  while( myAgora->GetStatus() != (int) DCDT_LETSWORK )
    Delay(20000);

  Status = RUNNING;

  // Execute our job
  Run();
  
  Close();
  
  myAgora->SetInactive( myID);

  // Terminate our thread
#ifndef USE_BOOST
  pthread_exit(NULL);
#endif

  return SUCCESS;
};

inline void DCDT_Member::Suspend() {

};

/*! \brief Execute our DoYourJob method periodically, if we are a periodical member or continuously.
 */
inline void DCDT_Member::Run() {

  DCDT_TIME start, last;

#ifdef NANOSLEEP_VERSION

  DCDT_TIME delta;
  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("Member %d now running", myID));

  pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, NULL);

  last = StartTime;
  while( (Status == RUNNING) && (myAgora->GetStatus() == DCDT_LETSWORK) )   {
    start = GetTime();
    CycleTime = start-last;
    last = start;

    DoYourJob();

    pthread_testcancel();

    ExecTime = GetTime() - start;

    // If this member is periodic we have to wait for the next period
    // We call always Wait because when using SCHED_RR the call
    // allow another thread to be scheduled
    if( Period ) {
      delta = Period - ExecTime;
      if ( delta > 0 ) {
	delta = delta - NANOBIAS;
	if ( delta < 0 ) {
	  delta = 1;
	}
      }
      else {
	CycleTimeExceded=true;
	delta = 1;
      }
    } else {
      delta = 1;
    }

    Wait ( delta );
  }

#endif

#ifdef SEQUENCER_VERSION

  DCDT_TIME before_delay, actual_delay, delta;

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("Member %d now running", myID));

  gettimeofday(&deadline, NULL);
  StartTime = tv2DCDT_TIME( deadline );

  // if this is a periodic member
  if (Period) {
    deadline = deadline + Period;
    Wait( deadline );
  }

  last = StartTime;
  while( (Status == RUNNING) && (myAgora->GetStatus() == DCDT_LETSWORK) )   {
    gettimeofday(&deadline, NULL);

    start = tv2DCDT_TIME( deadline);
    CycleTime = start-last;
    last = start;

    DoYourJob();

    pthread_testcancel();

    before_delay = GetTime();
    ExecTime = before_delay - start;

    if ( ExecTime > Period )
      CycleTimeExceded=true;

    if( Period ) {
      deadline = deadline + Period;
      Wait( deadline );
    } else {
      Wait ( 1 );
    }
  }
#endif

  TRC_PRINT( DCDT_TRC_MEMBER, TRC1, ("DCDT_Member::Run()- exit -  Member %d", myID ));
};

#ifndef SEQUENCER_VERSION
inline void DCDT_Member::Wait( DCDT_TIME sleep_time) {

  Delay( sleep_time );

};

inline void DCDT_Member::Wait( struct timeval deadline ) {
  struct timeval now;
  DCDT_TIME delay_time;
  gettimeofday(&now, NULL);

  delay_time = deadline - now;
  if (delay_time > 0)
    Delay( delay_time);
};

#else
inline void DCDT_Member::Wait( DCDT_TIME sleep_time) {
  mySequencer->Wait( sleep_time );
};

inline void DCDT_Member::Wait( struct timeval deadline ) {
  mySequencer->Wait( deadline );
};
#endif

#endif
