/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                    
**!!! PROJECT       : Middleware                                 
**!!!                                                                    
**!!! CLASS         : ModuleMember                     
**!!!
**!!! PACKAGE       : Middleware
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : ModuleMember.h
**!!!                                                                   
**!!! CREATION DATE : 09-07-2005         
**!!!                                                                   
**!!! AUTHOR        : Cristian Giussani
**!!!
**!!!                                                                   
**!!! OVERVIEW      : ModuleMember makes MIDDLEWARE transparent to 
**!!!	              expert modules (Task execution and string message 
**!!!                 passing)
**!!!                 MIDDLEWARE is DCDT.       
**!!!                                                                   
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
*/

#define EXPERTS_TIME 1

#ifndef __MODULE_MEMBER_H
#define __MODULE_MEMBER_H
/*----------------------------------------------------------------------
 *--------------------- SYSTEM  INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <DCDT_Member.h>
#include <DCDT_Agora.h>
#include <DCDT_Msg.h>

#include "stl.h"
/*----------------------------------------------------------------------
 *--------------------- PROJECT INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include "ModuleDispatch.h"

/*----------------------------------------------------------------------
 *-----------------------  DEFINE MACRO  -------------------------------
 *---------------------------------------------------------------------*/
const unsigned short ALL_TYPES=0;

/*----------------------------------------------------------------------
 *-----------------------  DEFINE TYPES  -------------------------------
 *---------------------------------------------------------------------*/

typedef enum {LOCAL,REMOTE} RequestType;

/*----------------------------------------------------------------------
 *-----------------------  DEFINE CLASS  -------------------------------
 *---------------------------------------------------------------------*/

/*! \brief This class is the interface that provided the basic functions for creating and senting messages.
 *
 *  The ModuleMember class is an abstract class ( since in contains three pure
 *  pure virtual methods which are inherited from the class @ref DCDT_Member )
 *  that gives the necessary interface to build messages.
 */
class ModuleMember: public DCDT_Member
{
  
  /*------------------- DEFINE CONSTRUCTOR/DESTRUCTOR -------------------*/
 protected:
  /** default Constructor **/
  ModuleMember( ModuleDispatch *Dispatch,
		unsigned int period=10000, string name="ModuleMember" );
  
  /** used to retet duty time in debug messages **/
  void ResetTimer();
  
  /** Destructor **/ 
  virtual  ~ModuleMember(void);
  
 public:
  // ModuleMember's sons have to implement these methods
  virtual void RunDuty(void)=0;
  virtual void RunInit(void)=0;
  virtual void RunClose(void)=0;

  /*********************************************
   **--          MESSAGE MANAGEMENT FUNCTIONS
   **-------------------------------------------
   **/
  
  /*! \brief Send the message pointed by \a Buffer through DCDT.
   *
   *  Create a new DCDT_Msg object using the user buffer, the
   *  calling process MUST dynamically allocate the buffer.
   *  If TCP_WARRANTY is specified as Warranty the message will be sent
   *  to other Agorae using TCP communication channels otherwise an UDP
   *  one will be used.
   *  DCDT will free the buffer when it will be useless.
   *
   *  \param Buffer - Pointer to the message
   *  \param Length - Message length
   *  \param Type - Message type
   *  \param Warranty - Message warranty ( UDP_WARRANTY or TCP_WARRANTY )
   *  \param Priority - Message priority
   *  \return RC_OK - Messages successfully sent
   *  \return RC_ERROR - Error while sending the message
   */
  int SendMessage(unsigned char * Buffer, int  Length,
		  int Type, int Warranty=UDP_WARRANTY, char Priority=0);
  
  /*! \brief Receive a message of type \a Type through DCDT ( it allocates the buffer )
   *
   *   NB: The buffer will be valid until the next call to a Receive... function.
   *
   *   \param Type - type of message to be received
   *   \param Buffer - pointer to the message received
   *   \param wait - If wait == true this method will wait until a message was found
   *   \return 0 - No new message available
   *   \return >0 - Size of the received message
   */
  int ReceiveMessage(int Type, unsigned char * &Buffer, bool wait=false );
  
  /*! \brief Get the last message received ( it allocates the buffer ).
   *
   *   NB: The buffer will be valid until the next call to a Receive... function.
   *
   *   \param Type - Type of message to be received
   *   \param Buffer - Pointer to the message received
   *   \param wait - If wait == true this method will wait until a message was found
   *   \return 0 - No new message available
   *   \return >0 - Size of the received message
   */
  int ReceiveLastMessage(int Type, unsigned char * &Buffer, bool wait=false );

  /*! \brief Clear all messages of type \a Type in queue.
   *
   *  \param Type - Type of messages ( default ALL_TYPES )
   *  \return RC_OK - Messages successfully deleted
   *  \return RC_ERROR - Error deleting messages
   */
  int ClearMessages(int Type=ALL_TYPES);
  
  /*! \brief Subscribe to the messages of type \a Type.
   *
   *  \param Type - Type of messages to subscribe to
   *  \param Flag - Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
   *  \return RC_OK - Subscription ok
   *  \return RC_ERROR - Error during subscription
   */
  int AddMessageRequest(int Type, RequestType Flag);
  
  /*! \brief Unsubscribe to the messages of type \a Type.
   *
   *  \param Type - Type of messages to subscribe to
   *  \param Flag - Subscribe to local ( LOCAL ) or remote ( REMOTE ) messages
   *  \return RC_OK - Subscription ok
   *  \return RC_ERROR - Error during subscription
   */
  int RemMessageRequest(int Type, RequestType Flag);
  
  /// Get Expert Id (Attributed at creation time by ModuleDispatch
  int GetId();
  
  /*! Shutdown
   */
  int Terminate(int uSec);
  
 protected:
  /// Module name
  string mModuleName;

  /// Dispatcher Reference
  ModuleDispatch* Dispatcher;
  
 private:
  /// Expert Identifier
  int MyId;

#ifdef EXPERTS_TIME
  unsigned int mExpertTime;
#endif
  
  /// Pointer to the incoming message, used in Receive methods as temporary variable
  const DCDT_Msg *IncomingMessage;
  
  /** Methods: **/
  // Redefinition of Middleware Init,Close and Main Process functions
  void DoYourJob(int wc);
  void Init(void);
  void Close(void);
};

#endif
