/*
  DEFINITIONS.H

  This file contains all global definitions
*/

#ifndef __DEFINITIONS_H
#define __DEFINITIONS_H

#include <p18f252.h> // Define PIC we are using

/*-----------------------------------
----------- DEFINITIONS -------------
-----------------------------------*/

// Having PWM @ 20Khz, pwm period is 50 us (0 - 500). So
// with pwm = 250 we are stopped
/*WARNING!!! If we change the PWM frequence, we have to change these values, because they are calculated for a specific PWM frequence.
We can find the formulae in the PIC datasheet, in PWM section.
*/
#define PWM_ZERO 250
#define PWM_MAX 250
#define PWM_MIN -250

//PIC pin
#define PIN_ENGINE_0	PORTBbits.RB0	//define pin to command engine0 //pin 21
#define DIR_ENGINE_0	PORTBbits.RB1  //input //pin 22
#define PIN_ENGINE_1	PORTBbits.RB2	//define pin to command engine1 //pin 23
#define DIR_ENGINE_1	PORTBbits.RB3  //input //pin 24
#define KICK_0 PORTBbits.RB4 //kicker 0 //pin 25
#define KICK_1 PORTBbits.RB5 //kicker 1 //pin 26

//engines definitions
#define ENGINE_0	0	//define engine0 value
#define ENGINE_1	1  //define engine1 value
#define BOTH_ENGINES 2 //define both engines: used when we want to set or use both engines

//engines status
#define ENGINE_ON	!ENGINE_OFF	//define engine status on
#define ENGINE_OFF 0	//define engine status off

//engines brake
#define ENGINE_GO	!ENGINE_BRAKE //define engine movement
#define ENGINE_BRAKE 0	//define engine brake

//engines direction
#define ENGINE_FORWARD !ENGINE_BACKWARD	//define engine forward direction
#define ENGINE_BACKWARD 0 	//define engine backward direction

//kickers definitions
#define KICKER_0 0 //define kicker 0
#define KICKER_1 1 //define kicker 1

//kickers status
#define KICKER_ON !KICKER_OFF  //kicker on	
#define KICKER_OFF 1 //kicker off

#define OFF 1
#define ON !OFF

#define FALSE 0
#define TRUE !FALSE

#define OK 1
#define ERROR !OK

#define DEBUG_DEFAULT 0
#define DEBUG_SMALL 1
#define DEBUG_MEDIUM 2
#define DEBUG_FULL 3

#define MAX_TEST 10

/*-----------------------------------
--------ERRORS CODE -----------------
-----------------------------------*/
#define UKNOWN_CMD 99
#define BAD_CMD 7
#define BAD_DEVC 1
#define BAD_DATA 2



/*-----------------------------------
--------EEPROM ADDRESS ----------
-----------------------------------*/

#define SIGN_EXP 1 //number of bytes reserved for the sign of the exponent
#define EXP 3 //number of bytes reserved for the exponent value. MAX exponent is 2^EXP
#define VALUE 12 //number of bytes reserved for the value. MAX number is 2^VALUE

#define POSITIVE_EXP 0 //exponent is positive
#define NEGATIVE_EXP 1 //exponent is negative

/*Here you can define pic internal eprom addresses for important variables storing.
The space needed for each variable is two eeprom memory cells.
The variable is stored in the follow way:
SIGN_EXP bits for sign of the exponent
EXP bits for the exponent
VALUE bits for the value
*/

/*Kick delay before re-kick*/
#define RE_KICK_DEL_ADDR 30

/*Cycles number of secur timer*/
#define NUM_CYC_ADDR 32

/*PID proportional constant for engine 0.*/
#define KP0_ADDR 10

/*PID integral constant for engine 0.*/
#define KI0_ADDR 12

/*PID derivative constant for engine 0.*/
#define KD0_ADDR 14

/*PID proportional constant for engine 1.*/
#define KP1_ADDR 16

/*PID integral constant for engine 1.*/
#define KI1_ADDR 18

/*PID derivative constant for engine 1.*/
#define KD1_ADDR 20

/*Sampling period. Sampling period is in milliseconds (ms)*/
#define CT_ADDR 22

/*Kick delay.*/
#define CK_ADDR 24

/*Max tick for engine 0.*/
#define MAXTC0_ADDR 26

/*Max tick for engine 1.*/
#define MAXTC1_ADDR 28

/*Tick to space for engine 0.*/
//#define TICSP0_ADDR 30

/*Tick to space for engine 1.*/
//#define TICSP1_ADDR 32


/*Definitions of adresses reserved for Fuzzy engine*/

/*Starting point (inclusive) of input table 1*/
#define MF1_B_ADDR 34

/*Size of input table 1*/
#define MF1_SIZE 101


/*Starting point (inclusive) of input table 2*/
#define MF2_B_ADDR 135

/*Size of input table 2*/
#define MF2_SIZE 101


/* Output Singleton */
#define SMALL_CONS 236
#define HIGH_CONS 237

/*Address of switch 0=Pid 1=Fuzzy*/
#define PID_XOR_FUZZY_ADDR 239


/*Starting point of rules*/
#define F_RULES_ADDR 241

/*Address of rules size*/
#define F_RULES_SIZE_ADDR 240


/*Address of maximum*/
#define F_INCR_MAX_ADDR 238


extern unsigned char lastIEaddress;
extern unsigned char lastIEvalue;


/*-----------------------------------
--------ENCODERS VARIABLES ----------
-----------------------------------*/

extern long total_tick_0; //engine 0 total tick
extern long total_tick_1; //engine 1 total tick
extern unsigned int encoder_read_0, encoder_read_1;
extern int setpoint_0, setpoint_1;

/*-----------------------------------
------------ CONTROL ----------------
-----------------------------------*/

extern unsigned char control_counter;
extern int current_speed_0, current_speed_1;
extern long already_read_ticks_0, already_read_ticks_1;
extern unsigned float kp_0, kp_1, kd_0, kd_1, ki_0, ki_1;
extern unsigned float ticks_to_pwm_0, ticks_to_pwm_1;
extern unsigned int max_tick_0, max_tick_1;
extern unsigned char debug;
extern unsigned char tuning;
extern unsigned float ticks_to_speed_0, ticks_to_speed_1;
extern unsigned float ticks_to_space_0, ticks_to_space_1;

extern unsigned char is_fuzzy;
extern unsigned char num_of_rules;
extern int f_incr_max;

extern unsigned char rules[256-F_RULES_ADDR];
extern int rules_conseguent[256-F_RULES_ADDR];
extern unsigned char rules_matching[256-F_RULES_ADDR];

/*-----------------------------------
------- CUMMUNICATION BUFFER --------
-----------------------------------*/
#define DIM_INPUT_BUFFER 22 //input buffer size
#define DIM_SERIAL_WRITE 12 //buffer used when writing to serial

#define VARIABLE_SIZE 16
#define VALUE_SIZE 16

extern char parsingInputBuffer[]; // Input buffer
extern unsigned char dataReady; //TRUE if command is ready to be execute
extern unsigned char byteReceived; //number of byte received
extern unsigned char Command; //command to execute
extern unsigned char Device; //device to control
extern int InputData0; //input data 0
extern int InputData1; //input data 1

extern int FuzzyData;
extern int FuzzyOut;

/*-----------------------------------
--------- OTHERS VARIABLES ----------
-----------------------------------*/
extern char string_temp[40];

#endif
