/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.awt.Button;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Color;
import javax.swing.JProgressBar;

/**
 * @author Bruschieri Stefano
 *
 */
public class ProgressDialog extends Dialog implements ActionListener {

	public JProgressBar progressBar;
	private Label label = new Label("Copying first table", Label.LEFT);
	private boolean stopped = false;
	private int processLength;
	private Button okButton, cancelButton;
	private String errorString;
	
	/**
	 * Creates a progress a dialog with a progress bar, a label and two
	 * buttons. 
	 * 
	 *  <ul>
	 *   <li>The progress bar communicates the progress of an event by displaying its percentage of completion</li>
	 *   <li>The label shows what the program is doing or, if any errors occurs, it shows some diagnostic information</li>
	 *   <li>Ok button is enable only if process ends (with or without errors)</li>
	 *   <li>Cancel button is enable only if the process is executing</li>
	 *  </ul>
	 * 
	 * @param parent Any frame.
	 * @param processLength Length of process we want observe.
	 */
	public ProgressDialog(Frame parent, int processLength){
		super(parent, "Progress Bar", false);
		this.processLength = processLength;
		this.stopped = false;
		
		progressBar = new JProgressBar(0, processLength);
		progressBar.setValue(0);
		progressBar.setStringPainted(true);
		
		Panel progressPanel = new Panel();
		progressPanel.setLayout(new GridLayout(2, 1));
		progressPanel.add(progressBar);
		progressPanel.add(this.label);
		add(progressPanel, "Center");
		
		Panel buttonPanel = new Panel();
		buttonPanel.setLayout(new GridLayout(1, 2));
		this.okButton = new Button("OK");
		okButton.setEnabled(false);
		okButton.addActionListener(this);
		buttonPanel.add(okButton, "Left");
		this.cancelButton = new Button("Cancel");
		cancelButton.addActionListener(this);
		buttonPanel.add(cancelButton, "Right");
		add(buttonPanel, "South");
		
		setSize(300, 120);
		setLocation(parent.getLocationOnScreen().x + 30, 
				    parent.getLocationOnScreen().y + 30);
		setVisible(true);
	}

	/**
	 * Returns true if the True if process has been stopped because any errors
	 * occurs or the user has stopped clicking on <code>Cancel</code> button. 
	 * By default, this is false.
	 * 
	 * @return whether the process has been stopped
	 */
	public boolean hasBeenStopped(){
		return stopped;
	}

	/**
	 * Sets text of label under the bar. It is used to show diagnostic information
	 * about occurred error or, simply, to write the name of the process in execution.
	 * 
	 * @param newText The text under the progress bar
	 */
	public void setText(String newText){
		this.label.setText(newText);
	}
	
	/**
	 * Stops the progress bar and writes some diagnostic information.
	 * It is used if any errors occurs.
	 * 
	 * @param errorString The cause of halt
	 */
	public void setCancelled(String errorString){
		this.errorString = errorString;
    	this.stopped = true;
	}
	
	/**
	 * Sets the progress bar to his maximum value.
	 * If any errors occur, the progress bar becomes red and the label
	 * shows some diagnostic information; if process ends without errors
	 * the progress bar becomes green.
	 */
	public void setCompleted(){
		// No error has occurred 
		if(!this.stopped){
			this.setText("Completed");
			this.progressBar.setValue(this.processLength);
			this.progressBar.setForeground(new Color(28,155,25));
			this.okButton.setEnabled(true);
			this.cancelButton.setEnabled(false);
		} else {
			// User has cancelled the execution of process
	    	this.progressBar.setValue(this.processLength);
	    	this.progressBar.setString("cancelled");
	    	this.label.setText(this.errorString);
	    	this.progressBar.setForeground(new Color(255,0,0));
			this.okButton.setEnabled(true);
			this.cancelButton.setEnabled(false);
		}
	}
	
    /**
    * Handles events from the OK button. When OK is pressed the dialog becomes
    * invisible, disposes of its self, and retruns.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();

    	// User has pressed button OK
    	if(cmd.equals("OK")){
        	setVisible(false);
        	dispose();
    	}
    	
        // User has pressed button Cancel
        if(cmd.equals("Cancel")){
        	this.setCancelled("Cancelled by user");
    	}
    }

}
