#ifndef MultiSerialCommunication_h
#define MultiSerialCommunication_h

#include <unistd.h>
#include <termios.h>
#include <string>
#include <sstream>
#include <fcntl.h>


const int DEVICE_NO = 2;

#include "LowLevelCommunication.h"

extern int getTime();
//#define MOTOR_DEBUG 1

using namespace std;

/**
 * This class is the interface for the serial communication towards the low-level 
 * controller.
 *
 * The BiSerialCommunication class is an abstract class that implements the 
 * LowLevelCommunication interface for two serial communication devices, and it
 * gives the methods to write and read data of some types on the serial channel.
 *
 * @short	Interface for the serial communication
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class MultiSerialCommunication : public LowLevelCommunication
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  MultiSerialCommunication();

  /**
   * Destructor
   */
  virtual ~MultiSerialCommunication();

  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //SerialCommunication(const SerialCommunication& from);
  //SerialCommunication& operator=(const SerialCommunication& from);

  // OPERATIONS
  
  /**
   * This method initializes the serial communication channel.
   */
  void InitCommunication();
  
  /**
   * This method closes the communication channel.
   */
  void CloseCommunication();

  //
  //  void ReadEncoders(){};

  /**
   * This method writes a character on the serial channel
   *
   * @param c		The character that must be written on the serial channel
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void Write(char c, short int device);

  /**
   * This method writes an integer on the serial channel
   *
   * @param n	The number that must be written on the serial channel
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void Write(int n, short int device);

  /**
   * This method writes a string on the serial channel
   *
   * @param s	The string that must be written on the serial channel
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void Write(string s, short int device);

  /**
   * This method reads an integer from the serial channel
   *
   * @param n	A pointer to the integer that has been read from the serial
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void Read(int* n, short int device);

  /**
   * This method reads a string representing an integer from the serial channel
   *
   * @param n	A pointer to the integer that has been read from the serial
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void ReadNumber(int* n, short int device);

  /**
   * This method reads a string representing a float from the serial channel
   *
   * @param n	A pointer to the float that has been read from the serial
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void ReadNumber(float* n, short int device);

  /**
   * This method reads a char from the serial channel
   *
   * @param c	A pointer to the char that has been read from the serial
   * @param device	The device number to which sending the data (values: 0 and 1)
   */
  void Read(char* c, short int device);

  /**
   * Clear the buffers of the devices
   */
  void Clear();

 protected:

  /**
   * The descriptor for the nth serial device
   */
  int fd[DEVICE_NO];

  /**
   * The transmission rate of the nth serial device
   */
  unsigned int mBaudRate[DEVICE_NO];

  /**
   * A string that identifies the nth serial device
   */
  char* mModemDevice[DEVICE_NO];

  /**
   * True if the first control device needs the output hardware flow control
   */
  bool mHardwareFlowControl[DEVICE_NO];

 private:
  struct termios oldtio[DEVICE_NO];
  struct termios newtio[DEVICE_NO];
};


inline void MultiSerialCommunication::Write(char c, short int device)
{
  if (device <= (DEVICE_NO-1))
     write(fd[device],&c,1);
  else
     fprintf(stderr,"Error: device %d not found",device);     
}


inline void MultiSerialCommunication::Write(string s, short int device)
{
#ifdef MOTOR_DEBUG
  printf("\nWriting: %s\n",s.data());
#endif
#ifdef TIME
  int time = getTime();
#endif
  if (device <= (DEVICE_NO-1))
     write(fd[device],s.c_str(),s.size());
  else
     fprintf(stderr,"Error: device %d not found",device);

#ifdef TIME
  time = getTime() - time;
  printf("\nWriting Time: %d\n",time); 
#endif
}


inline void MultiSerialCommunication::Write(int n,short int device)
{
  if (device <= (DEVICE_NO-1))
     write(fd[device],&n,sizeof(int));
  else
     fprintf(stderr,"Error: device %d not found",device);
}

inline void MultiSerialCommunication::Read(int* n, short int device)
{
  if (device <= (DEVICE_NO-1))
     read(fd[device],&n,sizeof(int));
  else
     fprintf(stderr,"Error: device %d not found",device);
}

inline void MultiSerialCommunication::ReadNumber(int* n, short int device)
{
  stringstream received_message;

  char c = '\0';
  int flag=0;
  int number;

   if (device <= (DEVICE_NO-1))
     {
       while (c != '\n' && c != ' ' && c != '\t' && !flag)
	 {
	   Read(&c, device);
#ifdef MOTOR_DEBUG
	   printf("\nReadingNumber: %c\n",c);
#endif     
	   if( (c >= '0' && c <= '9') || c == '-' || c == '+' )
	     received_message << c;
	   else if( c == '\n' || c == ' ' || c == '\t' )
	   { }
	   else
	     flag=1;       
	 }

       received_message >> number;
       if(flag || abs(number)>=5000)
	 {  
	   *n=0;
#ifdef MOTOR_DEBUG
	   printf("\nGot: %d\n",*n);
#endif      
	 }
       else
	 {  
	   *n = number;
#ifdef MOTOR_DEBUG
	   printf("\nGot: %d\n",*n);
#endif      
	 }
     }
   else
     fprintf(stderr,"Error: device %d not found",device);
}


inline void MultiSerialCommunication::ReadNumber(float* n, short int device)
{
  stringstream received_message;

  char c = '\0';
  if (device <= (DEVICE_NO-1))
    {
      
      while (c != '\n')
	{
#ifdef TIME
	  int time = getTime();
#endif
	  Read(&c,device);
#ifdef TIME
	  time = getTime() - time;
	  printf("\nReading Time: %d from device no.%d\n",time,device);
#endif
	  if ((int)c != 6)
	    {
#ifdef TIME
	      int time = getTime();
#endif
	      received_message << c;
#ifdef TIME
	  time = getTime() - time;
	  printf("\nReading Time to Buffer: %d\n",time);
#endif
	}
    }
#ifdef TIME
  int time = getTime();
#endif
  received_message >> *n;
#ifdef TIME
  time = getTime() - time;
  printf("\nWriting to float: %d\n",time);
#endif
  
#ifdef MOTOR_DEBUG
  printf("\nReading string: %s\n",received_message.str().data());    
  printf("\nReading: %f\n",*n);
#endif      
  }
  else
     fprintf(stderr,"Error: device %d not found",device);
}


inline void MultiSerialCommunication::Read(char* c, short int device)
{
  if (device <= (DEVICE_NO-1))
     read(fd[device],c,1);
  else
     fprintf(stderr,"Error: device %d not found",device);

#ifdef MOTOR_DEBUG
  printf("\nReading: %d\n",(int)*c);
#endif   
}

#endif
