#ifndef MessageWriter_h
#define MessageWriter_h

// SYSTEM INCLUDES
//
#include <assert.h>
#include <iostream>
#include <sstream>
#include <string>
#include <assert.h>

// LOCAL INCLUDES
//
#include "ModuleExpert.h"


/**
 * \brief This class is the interface that provided the basic functions for creating and
 * sending string messages.
 *
 * The MessageWriter class is an abstract class (since it contains three 
 * pure virtual methods which are inherited from the class @ref ModuleExpert)
 * that gives the necessary interface to build string messages. In order to
 * write a message, first of all we have to call the method @ref #NewMessage
 * specifying the number that identifies the message and a string that is
 * associated to the message content. Then we can fill the message using the 
 * "<<" operator of the member mCurrentMessage or exploiting the methods provided 
 * by derived classes. Once the message has been completed, we can send it
 * through the @ref #SendStringMessage by specifying the message identifier.
 *
 * @author	Marcello Restelli
 * @author      Cristian Giussani
 * @short	Interface for string messages
 */

class StringMessageManager : virtual public ModuleExpert
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   *
   * @param pDispatcher A pointer to the dispatcher
   * @param period	The time that should pass between two consecutive 
   *			activations of the expert (in microseconds)
   * @param module_name A string that is associated to this expert
   */
  StringMessageManager(ModuleDispatch* pDispatcher, long period, string module_name);

  /**
   * Destructor
   */
  virtual ~StringMessageManager();

 protected:

  /**
   * Must be invoked every time we want to build a new message. If we were 
   * writing a message, this method closes that message, puts it in the
   * message queue and create a new message.
   *
   * @param msg_id	The number that identifies the message
   * @param msg_type	A string that is associated to the message content
   *			(e.g., localization_data, vision_data)
   */
  void NewMessage(unsigned int msg_id, string msg_type);

  /**
   * Send all the queued messages that are associated to an identifier.
   * The current message is closed before the sending procedure.
   *
   * @param message	The identifier of the messages that must be sent
   */
  void SendStringMessage(unsigned int message);

  /**
   * Send all the messages contained in the queue.
   */
  void SendAllStringMessages();

  /**
   * Insert the end tag at the end of the message, and put it in the queue.
   */
  void CloseCurrentMessage();

  /**
   * Insert all the string messages of a certain type into the mailbox.
   *
   * @param message	The identifier of the message
   */
  void ReceiveStringMessage(unsigned short message);

  /**
   * Read the last message received of a certain type and return it.
   *
   * @param message     The identifier of the message
   */
  unsigned char *GetLastStringMessage( unsigned short message );

  /**
   * Insert all the string messages received into the mailbox.
   */
  void ReceiveAllStringMessages();

  /**
   * Return the message stored in the i-th position.
   *
   * @param num	The position that we want read
   * @return	A pointer to the message in the i-th position
   */
  unsigned char* GetStringMessage(unsigned int num);

  /**
   * Return the number of received messages stored in the mailbox.
   *
   * @return	The number of messages in the mailbox
   */
  unsigned int GetNumberOfReceivedMessages();

  /**
   * Erase all the message in the mailbox. After this method the mailbox is empty.
   */
  void CancelAllStringMessages();

  
  virtual void RunInit()=0;
  virtual void RunDuty()=0;
  virtual void RunClose()=0;
 
  stringstream mCurrentMessage;
  bool mIsMessageOpened;

  
 private:
  unsigned int mMessageId;
  map<unsigned int, vector<string> > mMessages;
  ModuleDispatch* mpDispatcher;
  vector<unsigned char*> mReceivedMessages;
};

#endif
